#include <stdio.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <getopt.h>

//#define DEBUG 1

#define PORT_REQUEST    65001
#define PORT_REPLY      65002
#define DETECT_TIMEOUT  3

struct dev {
	struct in_addr addr;
	char mac[18];
	char ipaddr[16];
	char netmask[16];
	int mtu;
	char gateway[16];
	char dhcp;
	char* nameserver;

	struct dev* next;
};

struct option longopts[] = {
	{"mac-address",		1,	0,	'a'},
	{"bind-address",	1,	0,	'b'},
	{"dhcp-mode",		1,	0,	'd'},
	{"gw-address",		1,	0,	'g'},
	{"ip-address",		1,	0,	'i'},
	{"mtu-size",		1,	0,	'm'},
	{"netmask",		1,	0,	'n'},
	{"dns-server",		1,	0,	'z'},
	{"set-params",		0,	0,	's'},
	{"help",		0,	0,	'h'},
	{"display",		0,	0,	'D'},
	{0,			0,	0,	0}
};

static struct dev* devs = 0;
static int num_devs = 0;
static struct dev* d;
static char saved = 0;

#ifdef DEBUG
# define D(...) printf("[debug] " __VA_ARGS__);
#else
# define D(...)
#endif

// Usage Messages //
char usage_short[] = "\nUsage:    bfdetect -b<ip_host> -c<nr> -i<ip_card> -n<netmask> -g<gateway> -d<on|off> -m<mtu> -z<dns> -s]\n \
	  Enter 'bfdetect -h' to see more info on parameters!\n\n";
char usage_long[] = "Usage:\n\n \
   bfdetect --bind-address <ip_host> [options to set parameters of card]\n\n \
      -b | --bind-address <ip_host>\n \
         Set the address to bind bfdetect to. This must be the address of an interface\n \
         installed to the computer bfdetect is executed on.\n\n \
   Set IP-Parameters of card:\n\n \
      -a | --mac-address <mac address of card>\n \
         MAC-address of the card to be configured\n\n \
      -i | --ip-address <ip_card>\n \
	 IP-address that you want to assign to your bero*fix device\n\n \
      -n | --nm-address <netmask>\n \
         Netmask that you want to assing to your bero*fix device\n\n \
      -g | --gw-address <gateway>\n \
         IP-address of the gateway your bero*fix device should use\n\n \
      -d | --dhcp-mode <on|off>\n \
         En- or disables dhcp-client of your bero*fix device\n\n \
      -m | --mtu-size <500-1500>\n \
         Sets the MTU size of your bero*fix device to a value between 500 and 1500\n\n \
      -z | --dns-server <nameserver|none>\n \
         Specifies the nameserver used by your bero*fix device. Use 'none' to disable.\n\n \
      -s | --set-params\n \
         Really write the given settings to your bero*fix device\n\n \
   Misc parameters:\n\n \
      -D | --display\n \
         Displays list of bero*fix cards found in current network.\n \
	 This Parameter can be used in conjunction with --bind-address.\n\n \
   Example:\n\n \
      bfdetect -b192.168.0.128 -a00:00:00:aa:bb:cc -i192.168.120.2 -n255.255.255.0 -g192.168.120.1 -s\n\n \
      Will set the card with the given MAC-address to:\n \
         IP:      192.168.120.2\n \
	  Netmask: 255.255.255.0\n \
	  Gateway: 192.168.120.1\n\n \
      To set a card's IP-address the netmask and gateway have to be configured also.\n \
      The parameters ip-address, nm-address and gw-address have always to be used together.\n\n";
// Usage Messages //

static char save_response (struct in_addr from, char* reply, uint8_t printout)
{
	char good_reply[32];

	if (d->addr.s_addr != from.s_addr)
		return 0;

	sprintf(good_reply, "ok hw=%s", d->mac);
	saved = !strncmp(good_reply, reply, strlen(good_reply));

	return 1;
}

static char device_found (struct in_addr from, char* reply, uint8_t printout)
{
	struct dev* new_dev;
	char* p;
	char* key;
	char* val;
	int len;

	D("Client %s answered: %s\n", inet_ntoa(from), reply);

	new_dev = calloc(1, sizeof(struct dev));
	new_dev->addr = from;

	while ((p = strsep(&reply, ";"))) {
		key = strsep(&p, "=");
		if (!key)
			continue;
		val = strsep(&p, "\r\n");
		if (!val)
			continue;
		len = strlen(val);
		if (!len)
			continue;

		D("KEY:%s VAL:%s\n", key, val);

		if (!strcmp(key, "hw")) {
			if (len == 17)
				strncpy(new_dev->mac, val, len);
		}
		else if (!strcmp(key, "ip")) {
			if (len <= 16)
				strncpy(new_dev->ipaddr, val, len);
		}
		else if (!strcmp(key, "nm")) {
			if (len <= 16)
				strncpy(new_dev->netmask, val, len);
		}
		else if (!strcmp(key, "mtu")) {
			new_dev->mtu = atoi(val);
		}
		else if (!strcmp(key, "gw")) {
			if (len <= 16)
				strncpy(new_dev->gateway, val, len);
		}
		else if (!strcmp(key, "dhcp")) {
			new_dev->dhcp = (*val == '1');
		}
		else if (!strcmp(key, "ns")) {
			if (new_dev->nameserver)
				free(new_dev->nameserver);
			new_dev->nameserver = strdup(val);
		}
	}

	if (!new_dev->mtu)
		new_dev->mtu = 1500;

	new_dev->next = devs;
	devs = new_dev;
	num_devs++;

	if (printout)
		printf("[%d] mac:%s ip:%s\n", num_devs, new_dev->mac, new_dev->ipaddr);

	return 0;
}

static void receiver (char (*callback) (struct in_addr from, char* reply, uint8_t printout), uint8_t printout)
{
	struct sockaddr_in sa_server, sa_client;
	socklen_t sa_client_len = sizeof(sa_client);
	int s, len;
	char reply[512];
	fd_set fds;
	struct timeval tv;

	s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (s < 0)
		return;

	memset(&sa_server, 0, sizeof(sa_server));
	sa_server.sin_family = AF_INET;
	sa_server.sin_addr.s_addr = htonl(INADDR_ANY);
	sa_server.sin_port = htons(PORT_REPLY);

	if (bind(s, (struct sockaddr *)&sa_server, sizeof(sa_server)) < 0)
		return;

	FD_ZERO(&fds);
	FD_SET(s, &fds);

	tv.tv_sec = DETECT_TIMEOUT;
	tv.tv_usec = 0;

	while (select(s + 1, &fds, 0, 0, &tv)) {
		len = recvfrom(s, reply, sizeof(reply) - 1, 0,
					   (struct sockaddr *)&sa_client,
					   &sa_client_len);
		if (len < 0)
			return;

		reply[len] = 0;
		if (callback(sa_client.sin_addr, reply, printout))
			break;
	}

	close(s);
}

static int send_request (struct in_addr from, struct in_addr to, char* request)
{
	struct sockaddr_in sa_request, src;
	int s, sent = 0, opt = 1;
	int len = strlen(request);

	s = socket(PF_INET, SOCK_DGRAM, IPPROTO_UDP);
	if (s < 0)
		return -1;
	
	setsockopt(s, SOL_SOCKET, SO_BROADCAST, &opt, sizeof(opt));

	memset(&src, 0, sizeof(src));
	src.sin_family = AF_INET;
	src.sin_addr = from;

	if ( bind(s, (struct sockaddr *)&src, sizeof(src)) != 0 )
		return -1;

	memset(&sa_request, 0, sizeof(sa_request));
	sa_request.sin_family = AF_INET;
	sa_request.sin_addr = to;
	sa_request.sin_port = htons(PORT_REQUEST);

	sent = sendto(s, request, len, 0,
   				  (struct sockaddr *)&sa_request,
				  sizeof(sa_request));
	if (sent != len)
		return -1;

	close(s);

	return 0;
}

static void get_all (void)
{
	int input;

	while (1) {
		input = getchar();
		if (input == EOF || input == '\n')
			break;
	}
}

static int get_int (char *msg)
{
	int input;

	do {
		printf("%s", msg);
		fflush(stdout);
		if (scanf("%u", &input) == 1) {
			get_all();
			return input;
		}
		while (1) {
			input = getchar();
			if (input == EOF || input == '\n')
				break;
		}
	} while (1);
}

static char get_char (char* msg, char* good)
{
	int input;

	do {
		printf("%s", msg);
		fflush(stdout);
		while (1) {
			input = getchar();
			if (input == EOF || input == '\n')
				break;
			if (strchr(good, input)) {
				get_all();
				return input;
			}
		}
	} while (1);
}

static char verify_mac (char *str_m) {

	int	i	= 0,
		a[6];

	if (sscanf(str_m, "%x:%x:%x:%x:%x:%x", &a[0], &a[1], &a[2], &a[3], &a[4], &a[5]) != 6)
		return(0);

	for (i = 0; i < 6; i++)
		if ((a[i] < 0) || (a[i] > 0xff))
			return(0);

	return(1);
}

static char verify_address (char* adr)
{
	int l, m, n, o;

	if (sscanf(adr, "%d.%d.%d.%d", &l, &m, &n, &o) != 4)
		return 0;

	if (l < 0 || m < 0 || n < 0 || o < 0 ||
		l > 0xff || m > 0xff || n > 0xff || o > 0xff)
		return 0;

	sprintf(adr, "%d.%d.%d.%d", l, m, n, o);
	return 1;
}

static void get_address (char* msg, char* buf)
{
	buf[15] = 0;
	do {
		printf("%s", msg);
		fflush(stdout);
		if ((scanf("%15s", buf) == 1) &&
			verify_address(buf)) {
			get_all();
			return;
		}
		get_all();
	} while (1);
}

static void get_string (char* msg, char** buf)
{
	char tmp[128] = "";
	do {
		printf("%s", msg);
		fflush(stdout);
		if (scanf("%127s", tmp) == 1) {
			get_all();
			if (*buf)
				free(*buf);
			*buf = strdup(tmp);
			return;
		}
		get_all();
	} while (1);
}

static char verify_config (void)
{
	struct in_addr host, net, gw;

	if (!inet_aton(d->ipaddr, &host)) {
		printf("Failed: ip address invalid!\n");
		return 0;
	}
	if (!inet_aton(d->netmask, &net)) {
		printf("Failed: netmask invalid!\n");
		return 0;
	}
	if (!inet_aton(d->gateway, &gw)) {
		printf("Failed: gateway address invalid!\n");
		return 0;
	}

	if ((host.s_addr & net.s_addr) !=
		(gw.s_addr & net.s_addr)) {
		printf("Failed: ip address and gateway are not part of the same network!\n");
		return 0;
	}

	return 1;
}

static int search_berofix (struct in_addr src, struct in_addr bcast, uint8_t printout) {
	
	if (send_request(src, bcast, "find bero.*"))
		return(-1);

	receiver(device_found, printout);

	if (!num_devs)
		return(0);

	return(1);
}

static int change_config (struct in_addr src, struct in_addr bcast, uint8_t printout) {
	
	char save_cmd[512];	

	if (!verify_config())
		return(0);
	
	if (snprintf(save_cmd, sizeof(save_cmd), "set hw=%s;ip=%s;nm=%s;mtu=%d;gw=%s;dhcp=%d%s%s;",
		d->mac, d->ipaddr, d->netmask, d->mtu, d->gateway, d->dhcp, d->nameserver && *d->nameserver ? ";ns=" : "",
		d->nameserver && *d->nameserver ? d->nameserver : "") >= sizeof(save_cmd)) {
			printf("Error preparing save message!\n");
			return(0);
	}

	if (send_request(src, bcast, save_cmd)) {
		printf("Error sending save request!\n");
		return(0);
	}

	receiver(save_response, printout);
	
	if (saved)
		printf("Configuration saved!\n");
	else
		printf("Failed saving configuration\n");
	
	return(1);
}

int main (int argc, char* argv[]) {
	const struct in_addr bcast = { .s_addr = htonl(INADDR_BROADCAST) };
        struct  in_addr src;
	int num, devnum, mtu;
	char action;
	char save_cmd[512];

	int c, longind, card_id, next_try;
	extern char *optarg;
	int mtu_size;
	int card_found;
	char src_ip[20], card_mac[20], card_ip[20], card_nm[20], card_gw[20], nameserver[20], dhcp_mode[3];
	int a_flag = 0, b_flag = 0, c_flag = 0, d_flag = 0, g_flag = 0, i_flag = 0,
	    m_flag = 0, n_flag = 0, z_flag = 0, s_flag = 0, typo_flag = 0, D_flag = 0;
	int search_res = 0;

	memset(&src, 0, sizeof(src));
	inet_aton("0.0.0.0", &src);
	
	if (argc >= 2) {
		while ((c = getopt_long(argc, argv, "a:b:d:g:i:m:n:g:z:shD", longopts, &longind)) != -1) {
			switch (c) {
				case 'a':					// mac of bf to configure
					strcpy(card_mac, optarg);
					if (verify_mac(card_mac))
						a_flag = 1;
					else
						typo_flag = 1;
					break;
				case 'b':					// bind to address
					strcpy(src_ip, optarg);
					b_flag = inet_aton(optarg, &src) ? 1 : 0;
					break;
				case 'd':					// dhcp [on|off]
					strcpy(dhcp_mode, optarg);
					if (!strcmp("on", dhcp_mode) || !strcmp("off", dhcp_mode))
						d_flag = 1;
					else
						typo_flag = 1;
					break;
				case 'g':					// gateway to set
					strcpy(card_gw, optarg);
					if (verify_address(card_gw))
						g_flag = 1;
					else
						typo_flag = 1;
					break;
				case 'i':					// ip-address to set
					strcpy(card_ip, optarg);
					if (verify_address(card_ip))
						i_flag = 1;
					else
						typo_flag = 1;
					break;
				case 'm':					// mtu-size
					mtu_size = atoi(optarg);
					if (mtu_size > 500 && mtu_size <= 1500)
						m_flag = 1;
					else
						typo_flag = 1;
					break;
				case 'n':					// netmask to set
					strcpy(card_nm, optarg);
					if (verify_address(card_nm))
						n_flag = 1;
					else
						typo_flag = 1;
					break;
				case 'z':					// dns-server
					strcpy(nameserver, optarg);
					if (verify_address(nameserver) || !strcmp("none", nameserver))
						z_flag = 1;
					else
						typo_flag = 1;
					break;
				case 's':					// save config
					s_flag = 1;
					break;
				case 'h':					// print help
					fprintf(stderr, "%s", usage_long);
					return(EXIT_SUCCESS);
				case 'D':
					D_flag = 1;
					break;
				case '?':
					switch (optopt) {
						case 'a':
						case 'b':
						case 'd':
						case 'g':
						case 'i':
						case 'm':
						case 'n':
						case 'z':
							printf("%s", usage_short);
					}
					return(EXIT_FAILURE);
			}
		}
	}

	if (typo_flag) {
		printf("Do you see the typo?\n");
		return(EXIT_FAILURE);
	}

	if (b_flag)
		printf("Scanning on %s for bero*fix devices...\n", src_ip);
	else
		printf("Scanning network for bero*fix devices...\n");

	search_res = search_berofix(src, bcast, (a_flag ? 0 : 1));
	if (search_res == -1) {
		printf("Error sending request!\n");
		return(EXIT_FAILURE);
	} else if (search_res == 0) {
		printf("None found.\n");
		return(EXIT_SUCCESS);
	}
	
	if (D_flag)
		return(EXIT_SUCCESS);
	
	if (a_flag && (num_devs > 0)) {

		card_found = -1;
		devnum = 0;
		for (d = devs, devnum = num_devs - devnum; devnum; d = d->next, --devnum) {
			if (!strcmp(d->mac, card_mac)) {
				card_found = 1;
				break;
			}
		}

		if (card_found < 0) {
			printf("\nNo bero*fix card with MAC %s found, leaving.\n", card_mac);
			return(1);
		}
		
		if (i_flag && n_flag && g_flag) {
			strcpy(d->ipaddr, card_ip);
			strcpy(d->netmask, card_nm);
			strcpy(d->gateway, card_gw);
		}
		
		if (d_flag)
			if (!strcmp("on", dhcp_mode))
				d->dhcp = 1;
			else if (!strcmp("off", dhcp_mode))
				d->dhcp = 0;
		
		if (z_flag)
			if (!strcmp("none", nameserver)) {
				free(d->nameserver);
				d->nameserver = 0;
			} else {
				strcpy(d->nameserver, nameserver);
			}

		if (m_flag)
			d->mtu = mtu_size;
		
		printf("\nbero*fix #%d (%s) will be set up to:\n\n", devnum, d->mac);
		printf("   IP address: %s\n", d->ipaddr);
		printf("   Netmask:    %s\n", d->netmask);
		printf("   Gateway:    %s\n", d->gateway);
		printf("   DHCP:       %s\n", (d->dhcp ? "on" : "off"));
		printf("   DNS server: %s\n", d->nameserver);
		printf("   MTU size:   %d\n\n", d->mtu);
		
		if (s_flag)
			return(change_config(src, bcast, 1));
		else
			printf("Add -s to your prior command to perform changes!\n");

		return(EXIT_SUCCESS);
	}

	if (num_devs > 1) {
		printf("\n");
		while (1) {
			devnum = get_int("Select a device: ");
			if (devnum > 0 && devnum <= num_devs)
				break;
		}
		for (d = devs, devnum = num_devs - devnum; devnum; d = d->next, --devnum);
	} else
		d = devs;

	while (1) {
		printf("\n");
		printf("Configuration of selected device:\n");
		printf("        mac: %s\n", d->mac);
		printf("         ip: %s\n", d->ipaddr);
		printf("    netmask: %s\n", d->netmask);
		printf("    gateway: %s\n", d->gateway);
		printf("       dhcp: %s\n", d->dhcp ? "yes" : "no");
		printf("        mtu: %d\n", d->mtu);
		printf(" nameserver: %s\n", d->nameserver && *d->nameserver ? d->nameserver : "(not set)");
		printf("\n");

		if (!d->dhcp) {
			printf("[i] change ip\n");
			printf("[n] change netmask\n");
			printf("[g] change gateway\n");
			printf("[d] switch dhcp on\n");
		} else
			printf("[d] switch dhcp off\n");
		printf("[t] change mtu\n");
		printf("[m] change nameserver\n");
		printf("[s] save and quit\n");
		printf("[q] quit\n");
		printf("\n");
		
		next_try = 1;
		do {
			action = get_char("What would you like to do? ", d->dhcp ? "dtmsq" : "ingdtmsq");

			switch (action) {
			case 'i':
				printf("\n");
				get_address("ip address: ", d->ipaddr);
				break;
			case 'n':
				printf("\n");
				get_address("netmask: ", d->netmask);
				break;
			case 'g':
				printf("\n");
				get_address("gateway: ", d->gateway);
				break;
			case 't':
				printf("\n");
				do {
					mtu = get_int("mtu: ");
					if (mtu >= 500 && mtu <= 1500) {
						d->mtu = mtu;
						break;
					}
					printf("mtu value out-of-range!\n");
				} while (1);
				break;
			case 'd':
				d->dhcp = !d->dhcp;
				break;
			case 'm':
				printf("\n");
				get_string("nameserver [or \"none\"]: ", &d->nameserver);
				if (d->nameserver && !strcmp(d->nameserver, "none")) {
					free(d->nameserver);
					d->nameserver = 0;
				}
				break;
			case 's':
				if (!change_config(src, bcast, 1))
					next_try = 0;
			case 'q':
				return 0;
			}
		} while (next_try == 1);
	}

	return 0;
}
